package es.uca.xmleye.formats;

/*
 Copyright (C) 2005-2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.io.File;
import java.io.FilenameFilter;
import java.io.FileNotFoundException;
import java.util.Locale;
import java.util.ResourceBundle;
import javax.swing.AbstractListModel;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Schema;
import javax.xml.XMLConstants;

/**
 * Repositorio de formatos de documento disponibles.
 * 
 * Un repositorio de formatos de documento no es más que un directorio con una
 * serie de ficheros con extensión .format. Cada uno de ellos contiene la
 * definición de un formato de documento que puede abrir XMLEye utilizando un
 * conversor.
 * 
 * Es posible definir un repositorio compuesto pasando varias rutas en vez de
 * una, combinando efectivamente todos los repositorios en uno:
 * <ul>
 * <li>En la carga: si dos repositorios contienen el mismo fichero, la entrada
 * del repositorio con mayor prioridad oculta a la otra.</li>
 * 
 * <li>Al guardar: toda entrada siempre se intentará guardar en el repositorio
 * con mayor prioridad, independientemente de que el formato ya exista en otro
 * repositorio.</li>
 * </ul>
 * 
 * @author Antonio García Domínguez
 * @version 1.0
 */
public class FormatDescriptorRepository extends AbstractListModel {
	/**
	 * Identificador para serialización
	 */
	private static final long serialVersionUID = -7757419290087998132L;
	private List<FormatDescriptor> entradas;
	private String[] rutas;

	final private static String XSD_FILENAME = "accepted-doc.xsd";
	private Schema formatDescriptorXSD;

	private static final ResourceBundle errRes;

	static {
		errRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.ErrorMessages",
				Locale.getDefault());
	}

	/**
	 * Crea un nuevo repositorio compuesto.
	 * 
	 * @param rutas
	 *            Rutas de repositorios simples en orden descendiente de
	 *            prioridad.
	 * @exception FormatDescriptorException
	 *                Hubo un problema al cargar alguno de los repositorios.
	 */
	public FormatDescriptorRepository(String[] rutas)
			throws FormatDescriptorException {
		cargar(rutas);
	}

	/**
	 * Crea un nuevo repositorio simple.
	 * 
	 * Método de conveniencia para
	 * <code>FormatDescriptorRepository(String[])</code>.
	 * 
	 * @param ruta
	 *            Ruta relativa al directorio con las descripciones de los
	 *            formatos.
	 * @exception FormatDescriptorException
	 *                Hubo un problema al cargar los formatos.
	 */
	public FormatDescriptorRepository(String ruta)
			throws FormatDescriptorException {
		cargar(new String[] { ruta });
	}

	/**
	 * Recarga los datos de las rutas empleadas.
	 * 
	 * Descarta todos los cambios que no se hayan guardado hasta el momento.
	 */
	public void recargar() throws FormatDescriptorException {
		cargar(rutas);
	}

	/** Guarda los cambios realizados en los formatos del repositorio. */
	public void guardar() throws FormatDescriptorException {
		for (FormatDescriptor t : entradas) {
			t.guardar();
		}
	}

	/**
	 * Devuelve el formato de documento que corresponda.
	 * 
	 * @param f
	 *            Fichero cuyo descriptor de formato se busca.
	 * @return FormatDescriptor que le corresponde, o <code>null</code> si
	 *         ninguno coincide.
	 */
	public FormatDescriptor getFormatoPorFich(File f) {
		return getFormatoPorRuta(f.getAbsolutePath());
	}

	/**
	 * Devuelve el formato de documento que corresponda.
	 * 
	 * @param ruta
	 *            Ruta al documento.
	 * @return FormatDescriptor de documento que le corresponde, o
	 *         <code>null</code> si ninguno coincide.
	 */
	public FormatDescriptor getFormatoPorRuta(String ruta) {
		for (FormatDescriptor t : entradas) {
			for (String ext : t.getExtensiones()) {
				if (ruta.matches(".*\\." + ext))
					return t;
			}
		}

		return null;
	}

	/** IMPLEMENTACION DE ListModel * */
	public int getSize() {
		return entradas.size();
	}

	public Object getElementAt(int index) {
		return entradas.get(index);
	}

	/** METODOS PROTEGIDOS/PAQUETE * */
	private void cargar(String[] rutas) throws FormatDescriptorException {
		this.rutas = rutas;
		Map<String, FormatDescriptor> mapaEntradas = new HashMap<String, FormatDescriptor>();

		for (String ruta : rutas) {
			final File dirBase = new File(ruta);
			for (String rutaFormato : dirBase.list(new FiltroFormato())) {
				final File fichFormato = new File(rutaFormato);
				final String clave = fichFormato.getName();
				if (mapaEntradas.containsKey(clave))
					continue;

				final File[] rutasFormato = new File[rutas.length];
				for (int i = 0; i < rutasFormato.length; ++i) {
					rutasFormato[i] = new File(rutas[i] + File.separator
							+ clave);
				}

				mapaEntradas.put(clave,
						new FormatDescriptor(this, rutasFormato));
			}
		}

		entradas = new ArrayList<FormatDescriptor>(mapaEntradas.values());
	}

	List<FormatDescriptor> getFormatos() {
		return entradas;
	}

	void nombreCambiado(FormatDescriptor cambiado) {
		int pos;
		for (pos = 0; pos < entradas.size() && cambiado != entradas.get(pos); ++pos)
			;

		if (pos < entradas.size()) {
			fireContentsChanged(this, pos, pos);
		}
	}

	Schema getSchema() throws FileNotFoundException {
		if (formatDescriptorXSD != null) {
			return formatDescriptorXSD;
		}

		for (String ruta : rutas) {
			File xsdFile = new File(ruta + File.separator + XSD_FILENAME);
			if (xsdFile.canRead()) {
				try {
					formatDescriptorXSD = SchemaFactory.newInstance(
							XMLConstants.W3C_XML_SCHEMA_NS_URI).newSchema(
							xsdFile);
					return formatDescriptorXSD;
				} catch (org.xml.sax.SAXException ex) {
					System.err.println(ex.getLocalizedMessage());
				}
			}
		}

		throw new FileNotFoundException(errRes
				.getString("format_xsd_not_found"));
	}

	/**
	 * Filtro por extensión .format
	 * 
	 * @author Antonio García Domínguez
	 * @version 1.0
	 */
	private static class FiltroFormato implements FilenameFilter {
		public boolean accept(File dir, String name) {
			return name.matches(".+\\.format");
		}
	}

}