package es.uca.xmleye.gui;

/*
  Copyright (C) 2005-2009 Antonio García Domínguez

  This program is free software; you can redistribute it
  and/or modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation;
  either version 2 of the License, or (at your option) any
  later version.

  This program is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE. See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public
  License along with this program; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA 02111-1307 USA

*/

import java.util.*;
import javax.swing.*;

import java.io.File;
import java.net.URI;
import java.lang.reflect.InvocationTargetException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;
import java.util.concurrent.Callable;

import es.uca.xmleye.ParsingException;
import es.uca.xmleye.formats.FormatDescriptorException;
import es.uca.xmleye.formats.FormatDescriptorRepository;
import es.uca.xmleye.prefs.*;
import es.uca.xmleye.presentation.WndFormatsPresentationModel;
import es.uca.xmleye.presentation.WndMainPresentationModel;
import es.uca.xmleye.xslt.StylesheetConstants;
import es.uca.xmleye.xslt.StylesheetRepository;
import es.uca.xmleye.xslt.StylesheetRepositoryException;
import es.uca.xmleye.xslt.TransformationException;

public class Application {
    /** Versión del visor. */
    public final static String VERSION = "1.2.3";

    // Referencia para comunicación con la UI
    private WndMain wndPrincipal;

    // Internacionalización de mensajes de error
    private final static ResourceBundle errRes;

    static {
        errRes = ResourceBundle.getBundle(
                                          "es.uca.xmleye.l10n.ErrorMessages",
                                          Locale.getDefault());
    }

    // Nombre de la variable de entorno a usar
    private final static String FORMATS_DIR_ENV_VAR = "XMLEYE_FORMATS_DIR";
    // Nombre de la variable de entorno a usar
    private final static String PREF_DIR_ENV_VAR = "XMLEYE_PREF_DIR";

    // Ruta relativa al repositorio global
    public final static String REPO_GLOBAL = "formats";
    // Fichero de preferencias por defecto
    private final static String PREFS_FILENAME = "xmleye.opt";
    // Ruta relativa al repositorio local
    private final static String REPO_LOCAL;

    private WndFormatsPresentationModel modPresentacionFormatos;
    private WndMainPresentationModel modPresentacionPrincipal;
    private FormatDescriptorRepository fdRepository;
    private PreferencesManager prefsManager;

    static {
        final String dirFormatos = System.getenv(FORMATS_DIR_ENV_VAR);
        REPO_LOCAL = dirFormatos != null ? dirFormatos : System
            .getProperty("user.dir");
    }

    /**
     * <p> Inicializa la aplicacion, ensamblando los componentes
     * principales entre sí.  </p>
     *
     * <ol>
     * <li>Carga el L&amp;F personalizado (el L&amp;F por defecto es poco
     * estético).</li>
     * <li>Crea los subcontroladores.</li>
     * <li>Muestra la ventana principal.</li>
     * </ol>
     *
     * @param docsPorAbrir
     *            Documentos que se desean abrir pasados desde la línea de
     *            órdenes.
     * @throws StylesheetRepositoryException
     */
    public Application(String[] docsPorAbrir)
        throws ExecutionException, InterruptedException,
               FormatDescriptorException, InvocationTargetException {
        // No existe otra forma sencilla de activar antialiasing en fuentes.
        // Ésta sólo funciona en J2SE 1.5, y no está documentada en el API
        // oficial.
        System.setProperty("swing.aatext", "true");

        // Use the SystemLookAndFeel by default, unless the
        // swing.defaultlaf system property has been set
        JFrame.setDefaultLookAndFeelDecorated(true);
        JDialog.setDefaultLookAndFeelDecorated(true);
        if (System.getProperty("swing.defaultlaf") == null) {
            try {
                UIManager.setLookAndFeel(
                    UIManager.getSystemLookAndFeelClassName());
            } catch (Exception e) {
                System.err.println(errRes.getString(
                    "err_lookandfeel_notfound"));
            }
        }

        // Cargamos las preferencias
        final String dirPrefs = System.getenv(PREF_DIR_ENV_VAR);
        final String rutaPrefs = dirPrefs != null ?
            (dirPrefs + File.separator + PREFS_FILENAME)
            : PREFS_FILENAME;
        prefsManager = new SerializedPreferencesManager(rutaPrefs);
        try {
            prefsManager.load();
        } catch (PreferencesException pEx) {
            MessageManagerSingleton.getInstancia().mostrarError(
                errRes.getString("err_loadpref_pex")
                + "\n"
                + pEx.getLocalizedMessage());
        }

        // First try to open the provided paths as URIs, and only if it fails
        // try and use normal file paths
        final ArrayList<File> fichPorAbrir = new ArrayList<File>();
        for (String ruta : docsPorAbrir) {
            try {
                fichPorAbrir.add(new File(new URI(ruta)));
            }
            catch (Exception ex) {
                try {
                    fichPorAbrir.add(new File(ruta));
                }
                catch (Exception innerEx) {
                    JOptionPane.showMessageDialog(
                        null,
                        String.format(errRes.getString("err_open_fail")
                                      ,ruta));
                }
            }
        }

        fdRepository = new FormatDescriptorRepository(new String[] {
                REPO_LOCAL, REPO_GLOBAL });

        FutureTask<Object> taskCreateUI = new FutureTask<Object>(
            new Callable<Object>() {
                public Object call() throws Exception {
                    modPresentacionFormatos
                        = new WndFormatsPresentationModel(fdRepository);
                    modPresentacionPrincipal
                        = new WndMainPresentationModel(
                            fdRepository,
                            prefsManager,
                            new StylesheetRepository(
                                StylesheetConstants.DIR_HOJAS));
                    wndPrincipal = new WndMain(Application.this,
                                               fichPorAbrir,
                                               modPresentacionPrincipal,
                                               modPresentacionFormatos);
                    return null;
                }
            });
        SwingUtilities.invokeAndWait(taskCreateUI);
        // Exceptions thrown during UI creation should
        // be rethrown here, if any
        taskCreateUI.get();

        MessageManagerSingleton.getInstancia().setCompPadre(wndPrincipal);
    }

    /**
     * Cierra el programa tras realizar todas las operaciones necesarias.
     */
    public void close() {
        try {
            prefsManager.save();
        } catch (PreferencesException prefEx) {
            MessageManagerSingleton.getInstancia().mostrarError(
                String.format(errRes.getString("err_savepref_pex")) + "\n"
                + prefEx.getLocalizedMessage());
            System.exit(1);
        }

        System.exit(0);
    }

    /**
     * Punto de inicio de la aplicación.
     */
    public static void main(String[] args) {
        try {
            new Application(args);
        } catch (Exception ex) {
            MessageManagerSingleton.getInstancia().mostrarError(
                String.format(errRes.getString("err_unhandled_init"), ex
                              .getLocalizedMessage()));
            ex.printStackTrace();
        }
    }

}
